#define PI 3.1415926536
#define time (iTime/10.)

struct ray_t{
    vec3 origin;
    vec3 direction;
};
struct sphere_t{
    vec3 origin;
    float radius;
};
   
    
// scattering coefficients at sea level (m)
const vec3 betaR = vec3(5.5e-6, 13.0e-6, 22.4e-6); // Rayleigh 
const vec3 betaM = vec3(21e-6); // Mie

// scale height (m)
// thickness of the atmosphere if its density were uniform
const float hR = 7994.0; // Rayleigh
const float hM = 1200.0; // Mie


float rayleigh_phase_func(float mu){
    return 3.*(1.+mu*mu)/(16.*PI);
}

//A beloved classic
const float g = 0.76;
float henyey_greenstein_phase_func(float mu)
{
	return
						(1. - g*g)
	/ //---------------------------------------------
		((4. * PI) * pow(1. + g*g - 2.*g*mu, 1.5));
}

//This function came from https://www.scratchapixel.com/lessons/procedural-generation-virtual-worlds/simulating-sky
//It seems to be an extension of the HG phase function though
//I don't actually use it anywhere
float mie_phase_func(float mu){
    return 3.*(1.-g*g)*(1.+mu*mu)/(8.*PI)/(2.+g*g)/pow(1.+g*g-2.*g*mu,1.5);
}
    
//I haven't looked into the specifics of this function yet, I just copied it straight from https://www.shadertoy.com/view/XtBXDz
bool intersects_sphere(in ray_t ray, in sphere_t sphere, inout float t0, inout float t1){
    vec3 rc = sphere.origin - ray.origin;
	float radius2 = sphere.radius * sphere.radius;
	float tca = dot(rc, ray.direction);
	float d2 = dot(rc, rc) - tca * tca;
	if (d2 > radius2) return false;
	float thc = sqrt(radius2 - d2);
	t0 = tca - thc;
	t1 = tca + thc;
	return true;
}


const float earth_radius = 6360e3; // (m)
const float atmosphere_radius = 6420e3; // (m)

//control the intensity and color of the light sources
const vec3 sun_power = vec3(20.0);
const vec3 moon_power = sun_power * vec3(0.2,0.2,0.35);

const int num_samples = 14;
const int num_samples_light = 3; //The original shader used 8, but I think you can get away with 3 (speedup will be substantial)

const sphere_t atmosphere = sphere_t(vec3(0.),atmosphere_radius);
const sphere_t earth = sphere_t(vec3(0.),earth_radius);

bool get_sun_light(in ray_t ray, inout float opticalDepthR, inout float opticalDepthM){
    float t0,t1;
    intersects_sphere(ray, atmosphere,t0,t1);
    
    float march_pos = 0.;
    float march_length = t1 / float(num_samples_light);
    
    for(int i=0; i<num_samples_light; i++){
        vec3 s = ray.origin + ray.direction*(march_pos + 0.5*march_length);
        float height = length(s)-earth_radius;
        opticalDepthR += exp(-height/hR) * march_length;
        opticalDepthM += exp(-height/hM) * march_length;
        
        march_pos += march_length;
    }
    return true;
}

vec3 get_incident_light(in ray_t ray){
    vec3 sun_dir = normalize(vec3(.0, sin(time),cos(time))); //I need a better positioning method than this
    
    
    if(iMouse.z>0.1){sun_dir = normalize(vec3(.0, sin(iMouse.x*2.*PI/iResolution.x - 1.),cos(iMouse.x*2.*PI/iResolution.x - 1.)));}//Mouse control instead of time
    
    vec3 moon_dir = -sun_dir; //moon is jus opposite of sun, but it doesn't need to be
    
    
    
    float t0,t1;
    if(!intersects_sphere(ray, atmosphere, t0, t1)){return vec3(0.); }
    
    float march_length = t1 / float(num_samples);
    
    float mu = dot(sun_dir, ray.direction);
    float muMoon = dot(moon_dir, ray.direction);
    
    float phaseR = rayleigh_phase_func(mu);
    float phaseM = henyey_greenstein_phase_func(mu);
    float phaseMoonR = rayleigh_phase_func(muMoon);
    float phaseMoonM = henyey_greenstein_phase_func(muMoon);
    
    
    //The accumulated extincion coefficients multiplied by the path length (optical depth = "average density")
    float opticalDepthR = 0.;
    float opticalDepthM = 0.;
    float opticalDepthMoonR = 0.;
    float opticalDepthMoonM = 0.;
    
    vec3 sumR = vec3(0.);
    vec3 sumM = vec3(0.);
    vec3 sumMoonM = vec3(0.);
    vec3 sumMoonR = vec3(0.);
    
    float march_pos = 0.;
    
    for(int i=0; i<num_samples; i++){
        vec3 s = ray.origin + ray.direction*(march_pos + 0.5*march_length); //sample middle of step
        float height = length(s) - earth_radius;
        
        //sample densities
        float hr = exp(-height / hR) * march_length;
        float hm = exp(-height / hM) * march_length;
        //Add to optical depth of primary ray - these variables will work for both the sun and moon
        opticalDepthR += hr;
        opticalDepthM += hm;
        
        //The moon is created by literally doing everything twice with extra "moon" variables
        //I'm too lazy to find a better solution
        ray_t light_ray = ray_t(s, sun_dir);
        ray_t moon_ray = ray_t(s,moon_dir);
        
        float opticalDepthLightR = 0.;
        float opticalDepthLightM = 0.;
        float opticalDepthLightMoonR = 0.;
        float opticalDepthLightMoonM = 0.;
        
        
        bool overground = get_sun_light(light_ray, opticalDepthLightR,  opticalDepthLightM);
        bool overgroundMoon = get_sun_light(moon_ray, opticalDepthLightMoonR, opticalDepthLightMoonM);
        
        if(overground || overgroundMoon){
            vec3 tau = betaR * (opticalDepthR + opticalDepthLightR) 
                + betaM * 1.1 * (opticalDepthM + opticalDepthLightM);
            
            vec3 tauMoon = betaR * (opticalDepthR + opticalDepthLightMoonR) 
                + betaM * 1.1 * (opticalDepthM + opticalDepthLightMoonM);

            vec3 attenuation = exp(-tau);
            vec3 attenuationMoon = exp(-tauMoon);
            
            sumR += hr * attenuation;
            sumM += hm * attenuation;
            sumMoonR += hr * attenuationMoon;
            sumMoonM += hm * attenuationMoon;
        }
        march_pos += march_length;
    }
    
    
    //Compute the final color by adding sun and moon lighting
    vec3 col = sun_power * (sumR * phaseR * betaR +sumM * phaseM * betaM) 
        + moon_power * (sumMoonR * phaseMoonR * betaR + sumMoonM * phaseMoonM * betaM); 
    
    //Draw circles around the celesteal bodies
    //I know if-statements are bad
    if(dot(ray.direction, sun_dir) > 0.9995)col *= (1.5,1.5,1.5);
    if(dot(ray.direction, moon_dir) > 0.9995)col *= (1.9,1.9,1.9);
    
    
    return col;
}

void mainImage( out vec4 fragColor, in vec2 fragCoord )
{
    vec2 uv = fragCoord/iResolution.xy *2. - 1.;
    uv *= vec2(16./9.,1.); //aspec correction
    vec3 cameraCenter = vec3(0.,earth_radius+1.,0.);
    if(iMouse.z>0.1){cameraCenter.y += 10e3 * iMouse.y/iResolution.y;}//Mouse control camera altitude
    float cameraLength = 2.;
    ray_t primary_ray = ray_t(cameraCenter, normalize(vec3(uv+vec2(0.,.5),cameraLength))); //I probably need to figure out camera transforms before making this better
    vec3 col = get_incident_light(primary_ray);
    col = pow(col*5.0 / (col*5.0+vec3(1)),vec3(2.2)); //Gamma correction from someone's comment in the original shader
    fragColor = vec4(col,1.);
}



